/*
 * Copyright (c) 2019, Jeffrey Lee
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>

#ifdef __riscos__
#include "kernel.h"
#include "swis.h"
#endif

#define VERSION 0.07
#define VDATE "13/04/19"

FILE *in,*out; /* Input and output streams */

uint32_t fgetw(FILE *f)
{
	return (uint32_t) fgetc(f) | (fgetc(f) << 8) | (fgetc(f) << 16) | (fgetc(f) << 24);
}

void fputw(uint32_t a,FILE *f)
{
	fputc(a & 255,f);
	fputc((a>>8) & 255,f);
	fputc((a>>16) & 255,f);
	fputc((a>>24) & 255,f);
}

void dohelp()
{
	/* Output help message */
	printf("nobrainer V%.2f %s Jeffrey Lee\n",VERSION,VDATE);
	printf("\
Usage: nobrainer [<options>] [<infile>] [<outfile>]\n\
Options:\n\
  -h, --help  Output this help text and exit\n\
  -l  Convert to/from Asylum level file (default) (H)\n\
  -b  Convert to/from Asylum tiled background sprite (H)\n\
  -s  Convert to/from FastSpr files (H)\n\
  -e  Convert to/from compressed border files\n\
  -m  Convert to/from a hammer file\n\
  -t  Convert to the Asylum format\n\
  -f  Convert from the Asylum format\n\
If infile or outfile are '-', or not specified, standard input/output will be\n\
used. This may not work on Windows. On RISC OS, use C-type redirection and not\n\
the OS redirection (i.e. < foo instead of { < foo }). Under RISC OS, -t/-f can\n\
be ommitted in order to automatically identify the conversion direction - if\n\
the input file is data then it will be assumed to be an Asylum file. Current\n\
output formats are sprite files (when converting from Asylum) and fastspr/data\n\
files (when converting to Asylum). Options marked (H) support automatic\n\
de-Hammering when converting from Asylum. At the moment hammer files can only\n\
be read, not generated\n");
}

/*

				INPUT FILE DE-HAMMERER

*/

int32_t ihmr_translate(int32_t i,int32_t l,int32_t f)
{
	uint32_t a,b;
	if (f == 0)
	{
		if (i > l)
			i = l; /* Stick it in the overrun space */
		return i;
	}
	a = i;
	b = a << 2;
	a = a - (l>>2);
	if ((a & 0x80000000) == 0)
	{
		b++;
		b=b-l;
	}
	a = a - (l>>2);
	if ((a & 0x80000000) == 0)
	{
		b++;
		b=b-l;
	}
	a = a - (l>>2);
	if ((a & 0x80000000) == 0)
	{
		b++;
		b=b-l;
	}
	if (b > l)
		b = l; /* overrun */
	return b;
}

int32_t ihmr_getc()
{
	static uint8_t *buff;
	static uint8_t tbuff[4];
	static int32_t bpos;
	static int32_t type = 0;
	static int32_t repeat;
	static int32_t offset,offset2;
	static int32_t shift;
	static int32_t tmp;
	static int32_t fmt;
	static int32_t len;
	static int32_t lenup;

	/* Operates on FILE *in by design */
	if (type == 0) /* Not initialised */
	{
		tbuff[0] = fgetc(in);
		tbuff[1] = fgetc(in);
		tbuff[2] = fgetc(in);
		tbuff[3] = fgetc(in);
		if ((tbuff[0] != 'H') || (tbuff[1] != 'm') || (tbuff[2] != 'r'))
		{
			/* Not hammer file */
			bpos = 0;
			type = 1; /* 'not hammer' flag */
			return tbuff[0];
		}
		/* Hammer file */
		lenup = len = fgetw(in);
		while (lenup & 3)
			lenup++;
		bpos = -1;
		fmt = tbuff[3];
		/* Now read entire file */
		buff = (uint8_t *) malloc(lenup+4); /* Add some overrun space */
		type = fgetc(in);
		while (type != 255)
		{
			if (type < 0x10)
			{
				/* Decode... */
				if (type < 15)
					repeat = type+2;
				else
					repeat = 256;
				offset = fgetc(in);
				while (repeat > 0)
				{
					/* Stick in buffer */
					bpos++;
					buff[ihmr_translate(bpos,lenup,fmt)] = offset;
					repeat--;
				}
			}
			else if (type < 0x20)
			{
				repeat = (type & 0xF)+1;
				while (repeat > 0)
				{
					offset = fgetc(in);
					/* Stick in buffer */
					bpos++;
					buff[ihmr_translate(bpos,lenup,fmt)] = offset;
					repeat--;
				}
			}
			else if (type < 0x40)
			{
				repeat = ((type & 0x1C)/4)+2;
				offset = bpos-((type & 0x3)*256 + fgetc(in));
				while (repeat > 0)
				{
					tmp = buff[ihmr_translate(offset,lenup,fmt)];
					offset--;
					/* Stick in buffer */
					bpos++;
					buff[ihmr_translate(bpos,lenup,fmt)] = tmp;
					repeat--;
				}
			}
			else if (type < 0x80)
			{
				shift = (type & 6)/2;
				if (shift >= 2)
					shift++;
				shift = 1 << shift;
				offset = bpos-(((type & 1)*256) + fgetc(in));
				repeat = ((type & 0x38)/8)+2;
				offset2 = 0;
				while (repeat > 0)
				{
					tmp = buff[ihmr_translate(offset+(offset2/4),lenup,fmt)];
					offset2+=shift;
					/* Stick in buffer */
					bpos++;
					buff[ihmr_translate(bpos,lenup,fmt)] = tmp;
					repeat--;
				}
			}
			else if (type < 0xff)
			{
				repeat = ((type & 0x78)/8)+2;
				offset = bpos - (((type & 7)*256)+fgetc(in));
				while (repeat > 0)
				{
					tmp = buff[ihmr_translate(offset,lenup,fmt)];
					offset++;
					/* Stick in buffer */
					bpos++;
					buff[ihmr_translate(bpos,lenup,fmt)] = tmp;
					repeat--;
				}
			}
			type = fgetc(in);
		}
		type = 2;
		bpos = 0;
	}
	if (type == 1)
	{
		/* Straight from file */
		if (bpos < 3) /* let in 0,1,2, which will output [1],[2],[3] */
		{
			bpos++;
			return tbuff[bpos];
		}
		return fgetc(in);
	}
	if (type == 2)
	{
		if (bpos < len)
			tmp = buff[bpos];
		else
			tmp = EOF;
		bpos++;
		return tmp;
	}
	printf("Bugger!\n");
	return EOF;
}

uint32_t ihmr_getw()
{
	uint32_t tmp;
	tmp = (uint32_t) ihmr_getc();
	tmp |= ((uint32_t) ihmr_getc()) << 8;
	tmp |= ((uint32_t) ihmr_getc()) << 16;
	tmp |= ((uint32_t) ihmr_getc()) << 24;
	return tmp;
}

/*

				LEVEL FILE CONVERTERS

*/

int32_t fromlevel()
{
	int32_t tmp,xs,ys,x,y;
	/* Read a level from in and output it to out... */
	ihmr_getw(); /* Ignore any values that are unknown */
	xs = ihmr_getw();
	ys = ihmr_getw();
	ihmr_getw();
	ihmr_getw();
	ihmr_getw();
	ihmr_getw();
	ihmr_getw();
	/* Now, can I work out how to output it straight to a sprite... */
	/* ...yes, because the sprites are in the same pixel order as the files :) */
	fputw(1,out); /* Number of sprites */
	fputw(16,out); /* Offset to 1st sprite (+4 as if the area is in memory, where the base would be a further 4 bytes back) */
	fputw(16+44+(xs*ys),out); /* First free word */
	fputw(44+(xs*ys),out); /* Offset to next sprite */
	fputs("asylumlevel",out); /* 11 bytes */
	fputc(0,out); /* 12 bytes */
	fputw((xs/4)-1,out); /* Width */
	fputw(ys-1,out); /* Height */
	fputw(0,out); /* 1st bit used per row */
	fputw(31,out); /* last bit used per row */
	fputw(44,out); /* Offset to image */
	fputw(44,out); /* Offset to mask (=image if no mask) */
	fputw(13,out); /* Mode */
	/* Data... */
	for (y=0;y<ys;y++)
		for (x=0;x<xs;x+=4)
		{
			tmp = ihmr_getw();
			fputw(tmp,out); /* Direct transfer word by word */
		}
	/* Yay! */
	return 0xff9; /* Sprite format */
}

int32_t tolevel()
{
	int32_t tmp,xs,ys,x,y;
	char copy[32];
	/* Convert from sprite to level */
	/* Sprite must be 1st in file, and have no wastage... */
	fgetw(in); /* Number of sprites */
	tmp = fgetw(in)-12; /* Offset to first sprite... */
	while (tmp+16 > 0)
	{
		fgetw(in); /* Skip to sprite x size */
		tmp-=4;
	}
	xs = (fgetw(in)+1)*4;
	ys = fgetw(in)+1;
	fgetw(in); fgetw(in); /* first & last bits used */
	tmp = fgetw(in)-36; /* Offset to sprite */
	while (tmp > 0)
	{
		fgetw(in); /* Skip to sprite data */
		tmp-=4;
	}
	/* Now output */
	fputw(0,out);
	fputw(xs,out);
	fputw(ys,out);
	sprintf(copy,"nobrainer V%.2f",VERSION);
	if (strlen(copy) > 20)
		copy[20] = 0; /* Doh! */
	fputs(copy,out);
	tmp = 20-strlen(copy);
	while (tmp > 0)
	{
		fputc(0,out);
		tmp--;
	}
	/* Now data */
	for (y=0;y<ys;y++)
		for (x=0;x<xs;x+=4)
		{
			tmp = fgetw(in);
			fputw(tmp,out); /* Direct transfer word by word */
		}
	return 0xffd; /* Data */
}

/*

				BACKFILE CONVERTERS

*/

int32_t frombackfile()
{
	int32_t tmp,xs,ys,x,y;
	/* Read a level from in and output it to out... */
	xs = 48;
	ys = 32;
	fputw(1,out); /* Number of sprites */
	fputw(16,out); /* Offset to 1st sprite (+4 as if the area is in memory, where the base would be a further 4 bytes back) */
	fputw(16+44+(xs*ys),out); /* First free word */
	fputw(44+(xs*ys),out); /* Offset to next sprite */
	fputs("backfile",out); /* 8 bytes */
	fputw(0,out); /* 12 bytes */
	fputw((xs/4)-1,out); /* Width */
	fputw(ys-1,out); /* Height */
	fputw(0,out); /* 1st bit used per row */
	fputw(31,out); /* last bit used per row */
	fputw(44,out); /* Offset to image */
	fputw(44,out); /* Offset to mask (=image if no mask) */
	fputw(13,out); /* Mode */
	/* Data... */
	for (y=0;y<ys;y++)
		for (x=0;x<xs;x+=4)
		{
			tmp = ihmr_getw();
			fputw(tmp,out); /* Direct transfer word by word */
		}
	/* Yay! */
	return 0xff9; /* Sprite format */
}

int32_t tobackfile()
{
	int32_t tmp,xs,ys,x,y;
	/* Convert from sprite to level */
	/* Sprite must be 1st in file, and have no wastage... */
	fgetw(in); /* Number of sprites */
	tmp = fgetw(in)-12; /* Offset to first sprite... */
	while (tmp+16 > 0)
	{
		fgetw(in); /* Skip to sprite x size */
		tmp-=4;
	}
	xs = (fgetw(in)+1)*4;
	ys = fgetw(in)+1;
	fgetw(in); fgetw(in); /* first & last bits used */
	tmp = fgetw(in)-36; /* Offset to sprite */
	while (tmp > 0)
	{
		fgetw(in); /* Skip to sprite data */
		tmp-=4;
	}
	/* Now output */
	for (y=0;y<ys;y++)
		for (x=0;x<xs;x+=4)
		{
			tmp = fgetw(in);
			fputw(tmp,out); /* Direct transfer word by word */
		}
	return 0xffd; /* Data */
}

/*

				FASTSPR CONVERTERS

*/

typedef struct {
	int32_t x,y;
	int32_t cx,cy;
	int32_t mode;
	uint8_t *sprite;
	uint8_t *mask;
} str_spr;

int32_t fromfastspr()
{
	/* Yay! Convert FastSpr files! */
	uint32_t utmp;
	int32_t tmp,sprites,*sprloc; /* Need to keep track on where sprites are since they may not be in order */
	int32_t xs,ys,*rowlocs;
	int32_t pos,closest,rlen;
	int32_t x,y;
	int32_t ndef,ab; /* Number of sprites defined */
	char name[16];
	str_spr *dat; /* Store data in RAM before outputting */
	utmp = ihmr_getw();
	if (utmp == 0x31505346)
		tmp = 1;
	else if (utmp == 0x32505346)
		tmp = 2;
	else
		tmp = 0;
	if (tmp == 0)
	{
		fprintf(stderr,"Unrecognised file format!\n");
		return 0;
	}
	if (tmp == 2)
	{
		fprintf(stderr,"Sorry, unable to read FSP2 format files\n");
		return 0;
	}
	/* Else format 1... */
	ihmr_getw(); ihmr_getw(); /* Skip the blanks */
	sprites = ihmr_getw();
	sprloc = (int32_t *) malloc(4*(sprites+1));
	for (tmp=0;tmp<sprites;tmp++)
		sprloc[tmp] = ihmr_getw();
	sprloc[sprites] = 0x7FFFFFFF; /* Dummy entry */
	dat = (str_spr *) malloc(sizeof(str_spr)*sprites);
	ndef = 0;
	memset(dat,0,sprites*sizeof(str_spr));
	/* Now work out where we are... */
	pos = 16+(sprites*4);
	ab = 0; /* Abort reading - EOF has been reached */
	/* So go through until we reach a sprite address... */
	do {
		closest = sprites; /* Close to no sprite */
		for (tmp=0;tmp<sprites;tmp++)
			if ((sprloc[tmp] >= pos) && (sprloc[tmp] < sprloc[closest]))
				closest = tmp;
		if (closest != sprites)
		{
			/* Continue until we reach the sprite */
			while (pos != sprloc[closest])
			{
				if (ihmr_getc() == EOF)
				{
					ab = 1;
					break; /* Escape from while */
				}
				pos++;
			}
			if (ab)
				break; /* Escape to main loop */
			/* Now read the sprite */
			xs = ihmr_getc(); /* X size */
			if (xs == EOF)
			{
				ab = 1;
				break; /* Can go up to sprite, but can't read it */
			}
			ndef++; /* Valid sprite */
			if (xs == 0)
				xs = 256;
			ys = ihmr_getc(); /* Y size */
			if (ys == 0)
				ys = 256;
			/* Store it in the array */
			dat[closest].x = xs;
			dat[closest].y = ys;
			dat[closest].sprite = (uint8_t *) malloc(xs*ys);
			dat[closest].mask = (uint8_t *) malloc(xs*ys);
			memset(dat[closest].sprite,0,xs*ys);
			memset(dat[closest].mask,0,xs*ys);
			dat[closest].cx = ihmr_getc(); /* Get centre locations */
			dat[closest].cy = ihmr_getc();
			/* Now the row lookup table */
			rowlocs = (int32_t *) malloc(4*(ys+1));
			for (tmp=0;tmp<ys+1;tmp++)
				rowlocs[tmp] = ihmr_getw(); /* Read in all the entries */
			/* Now the rows... */
			for (y=0;y<ys;y++)
			{
				rlen = rowlocs[y+1] - rowlocs[y]; /* Work out how much data there is for this row */
				while (rlen > 0)
				{
					utmp = ihmr_getw();
					if ((y > 0) && (dat[closest].mode == 0))
					{
						x = (int32_t) (utmp >> 8);
						x = x/320;
						if (x == y)
							dat[closest].mode = 13;
						else
							dat[closest].mode = 28; /* Assume it's 640 */
					}
					x = (int32_t) (utmp >> 8); /* Get position */
					x = x % 320; /* MOD with 320 to get real x pos */
					/* write pixel */
					dat[closest].sprite[x+(xs*y)] = utmp & 255;
					dat[closest].mask[x+(xs*y)] = 255;
					rlen-=4;
				}
			}
			if (dat[closest].mode == 0)
				dat[closest].mode = 13; /* Trap any fully masked sprites */
			/* End of sprite */
			pos = pos + rowlocs[ys]; /* Last row gives the sprite length */
			free(rowlocs);
		}
	} while ((closest != sprites) && (ab == 0));
	/* Finished, so output to disc... */
	fputw(ndef,out); /* Number of sprites */
	fputw(16,out); /* Offset to first sprite */
	pos = 16;
	for (tmp=0;tmp<sprites;tmp++)
		if (dat[tmp].x != 0)
			pos+=44+(((dat[tmp].x+3) & 0xFFC)*dat[tmp].y*2);
	fputw(pos,out); /* First free word - after all the sprites */
	for (tmp=0;tmp<sprites;tmp++)
	{
		if (dat[tmp].x != 0)
		{
			fputw(44+(((dat[tmp].x+3) & 0xFFC)*dat[tmp].y*2),out); /* Offset to next sprite */
			sprintf(name,"%" PRId32 "_%02" PRIx32 "%02" PRIx32,tmp,dat[tmp].cx,dat[tmp].cy); /* Output the centre pos in the sprite name */
			fputs(name,out); /* Output name */
			x = strlen(name);
			while (x < 12)
			{
				fputc(0,out);
				x++;
			}
			fputw((((dat[tmp].x+3) & 0xFFC)/4)-1,out); /* Width */
			fputw(dat[tmp].y-1,out); /* Height */
			fputw(0,out); /* 1st bit used per row */
			x = (((dat[tmp].x & 3)+4)*8)-1;
			fputw(x & 0x1F,out); /* Last bit used per row */
			fputw(44,out); /* Offset to image */
			fputw(44+(((dat[tmp].x+3) & 0xFFC)*dat[tmp].y),out); /* Offset to mask */
			fputw(dat[tmp].mode,out); /* Mode */
			for (y=0;y<dat[tmp].y;y++) /* Sprite */
			{
				for (x=0;x<dat[tmp].x;x++)
					fputc(dat[tmp].sprite[x+(y*dat[tmp].x)],out);
				while (x & 3)
				{
					fputc(0,out);
					x++;
				}
			}
			for (y=0;y<dat[tmp].y;y++) /* Mask */
			{
				for (x=0;x<dat[tmp].x;x++)
					fputc(dat[tmp].mask[x+(y*dat[tmp].x)],out);
				while (x & 3)
				{
					fputc(0,out);
					x++;
				}
			}
		}
	}
	/* Woohoo! */
	return 0xff9;
}


typedef struct {
	int32_t x,y,xc,yc,mode,num,numpix;
	uint8_t *sprite;
	uint8_t *mask;
} str_spr2;

int32_t tofastspr()
{
	/* Convert from sprite to fastspr... */
	/* Only FSP1 supported at the moment */
	/* Need to read in each sprite from the file and store it in mem */
	/* Then work out the max number and output the data from there */
	uint32_t utmp;
	int32_t tmp,sprites;
	str_spr2 *dat;
	int32_t pos,spr,tpos;
	int32_t ipos,mpos;
	char name[16];
	int32_t x,y;
	int32_t *sprord;
	sprites = fgetw(in); /* Get number of sprites */
	dat = (str_spr2 *) malloc(sizeof(str_spr2)*sprites);
	memset(dat,0,sprites*sizeof(str_spr2));
	tpos = fgetw(in); /* Offset to first sprite */
	pos = 12;
	for (spr=0;spr<sprites;spr++)
	{
		while (pos < tpos) /* Get to sprite */
		{
			fgetw(in);
			pos+=4;
		}
		/* Now read */
		tpos = fgetw(in); /* Next sprite */
		pos = 0;
		/* Get name */
		for (tmp=0;tmp<12;tmp++)
			name[tmp] = fgetc(in); /* Read name */
		/* Now process name */
		if (name[strlen(name)-5] != '_')
		{
			fprintf(stderr,"Error: '%s' is not a valid sprite name for fastspr conversion!\n",name);
			return 0;
		}
		sscanf(name,"%" SCNd32 "_%2" SCNx32 "%2" SCNx32,&dat[spr].num,&dat[spr].xc,&dat[spr].yc);
		dat[spr].x = fgetw(in)*4;
		dat[spr].y = fgetw(in)+1;
		fgetw(in); /* Ignore 1st bit per row */
		tmp = fgetw(in); /* But get last bit... */
		dat[spr].x+=(tmp+1)/8; /* Add the extra bytes */
		ipos = fgetw(in); /* Image offset */
		mpos = fgetw(in); /* Mask offset */
		tmp = fgetw(in); /* Mode */
		if (tmp == 13)
			dat[spr].mode = 320; /* Bytes per row */
		else
			dat[spr].mode = 640;
		pos = 44;
		dat[spr].sprite = (uint8_t *) malloc(dat[spr].x*dat[spr].y);
		dat[spr].mask = (uint8_t *) malloc(dat[spr].x*dat[spr].y);
		while (pos < ipos) /* Assume that image comes before mask, as the PRMs say it should */
		{
			fgetw(in);
			pos+=4;
		}
		if (pos == ipos)
		{
			/* Read image */
			for (y=0;y<dat[spr].y;y++)
			{
				for (x=0;x<dat[spr].x;x++)
					dat[spr].sprite[x+(y*dat[spr].x)] = fgetc(in);
				while (x & 3)
				{
					fgetc(in);
					x++;
					pos++; /* The calculation below doesn't take this into account */
				}
			}
			pos+=dat[spr].x*dat[spr].y;
		}
		if (ipos == mpos)
			memset(dat[tmp].mask,255,dat[spr].x*dat[spr].y);
		else
		{
			while (pos < mpos)
			{
				fgetw(in);
				pos+=4;
			}
			for (y=0;y<dat[spr].y;y++)
			{
				for (x=0;x<dat[spr].x;x++)
					dat[spr].mask[x+(y*dat[spr].x)] = fgetc(in);
				while (x & 3)
				{
					fgetc(in);
					x++;
					pos++;
				}
			}
			pos+=dat[spr].x*dat[spr].y;
		}
		for (y=0;y<dat[spr].x*dat[spr].y;y++)
			if (dat[spr].mask[y] == 255)
				dat[spr].numpix++;
		/* Read sprite, so on to next */
	}
	/* Finished, so now output the fastspr file... */
	fputw(0x31505346,out); /* 'FSP1' */
	fputw(0,out);
	fputw(0,out);
	/* Calculate highest sprite number */
	x = -1;
	for (spr=0;spr<sprites;spr++)
	{
		if (dat[spr].num > x)
			x = dat[spr].num;
	}
	fputw(x+1,out); /* Number of sprites */
	pos = 20+(x*4); /* Location of end of table */
	sprord = (int32_t *) malloc(4*sprites);
	tmp = 0;
	for (y=0;y<=x;y++) /* Output lookup table... */
	{
		tpos = 0; /* Not found an entry yet */
		for (spr=0;spr<sprites;spr++)
		{
			if (dat[spr].num == y)
			{
				tpos = 4+(dat[spr].y*4)+4+(dat[spr].numpix*4); /* Info word, row lookup, extra row lookup, pixel data */
				sprord[tmp] = spr; /* Lookup table for which order they should be output */
				tmp++;
			}
		}
		if (tpos != 0)
			fputw(pos,out); /* Output address */
		else
			fputw(0,out); /* No sprite. fspconv appears to output out of bounds addresses as well... */
		pos+=tpos;
	}
	/* Now output sprites */
	for (spr=0;spr<sprites;spr++)
	{
		tmp = sprord[spr];
		fputc(dat[tmp].x,out); /* X size */
		fputc(dat[tmp].y,out); /* Y size */
		fputc(dat[tmp].xc,out); /* X centre */
		fputc(dat[tmp].yc,out); /* Y centre */
		/* Now row lookup table... */
		pos = 4+(dat[tmp].y*4)+4; /* Info word, row lookup, extra row lookup */
		for (y=0;y<dat[tmp].y;y++)
		{
			fputw(pos,out);
			for (x=0;x<dat[tmp].x;x++)
				if (dat[tmp].mask[x+(dat[tmp].x*y)] == 255)
					pos+=4;
		}
		fputw(pos,out); /* Extra entry */
		/* Now the data */
		for (y=0;y<dat[tmp].y;y++)
		{
			for (x=0;x<dat[tmp].x;x++)
				if (dat[tmp].mask[x+(dat[tmp].x*y)] == 255)
				{
					utmp = dat[tmp].sprite[x+(dat[tmp].x*y)] + (x + dat[tmp].mode*y)*256;
					fputw(utmp,out);
				}
		}
	}
	/* Yay! */
	return 0x113;
}

/*

			COMPRESSED 'EDGE FILE' CONVERTERS

*/

int32_t fromedge()
{
	uint32_t utmp;
	int32_t tmp,count;
	for (tmp=0;tmp<0x44;tmp+=4)
		utmp = fgetw(in); /* Ignore header */
	fputw(1,out); /* Number of sprites */
	fputw(16,out); /* Offset to first */
	fputw(16+44+(320*256),out); /* Offset to first free word */
	fputw(44+(320*256),out); /* Offset to next sprite */
	fputs("edgesprite",out);
	fputc(0,out);
	fputc(0,out);
	fputw((320/4)-1,out); /* Width */
	fputw(255,out); /* Height */
	fputw(0,out); /* 1st bit per row */
	fputw(31,out); /* Last bit per row */
	fputw(44,out); /* Offset to image */
	fputw(44,out); /* Offset to mask */
	fputw(13,out); /* Mode */
	utmp = 0;
	while (utmp<320*256)
	{
		count = fgetc(in);
		if (count >= 0x80)
		{
			/* Length of a single colour */
			count = count-126; /* -128+2 */
			tmp = fgetc(in);
			while ((count > 0) && (utmp<320*256))
			{
				fputc(tmp,out);
				count--;
				utmp++;
			}
		}
		else
		{
			count = count+1;
			while ((count > 0) && (utmp<320*256))
			{
				fputc(fgetc(in),out);
				count--;
				utmp++;
			}
		}
	}
	return 0xff9;
}

int32_t toedge()
{
	int32_t tmp;
	char buff[130]; /* Output buffer to store data */
	int32_t bpos; /* Position in buffer */
	int32_t count;
	sprintf(buff,"nobrainer V%.2f",VERSION);
	bpos = 0;
	for (count=0;count<0x44;count++) /* Output header */
	{
		fputc(buff[bpos],out);
		if (buff[bpos])
			bpos++;
	}
	/* Locate at start of sprite... */
	fgetw(in); /* Number of sprites */
	tmp = fgetw(in)-12; /* Offset to first sprite... */
	while (tmp+16 > 0)
	{
		fgetw(in); /* Skip to sprite x size */
		tmp-=4;
	}
	fgetw(in); fgetw(in); /* Assume size is correct */
	fgetw(in); fgetw(in); /* first & last bits used */
	tmp = fgetw(in)-36; /* Offset to sprite */
	while (tmp > 0)
	{
		fgetw(in); /* Skip to sprite data */
		tmp-=4;
	}
	/* No output */
	bpos = 0;
	tmp = 0; /* uncompressed section */
	for (count=0;count<320*256;count++)
	{
		if (tmp == 0)
		{
			if (bpos == 128) /* Full? */
			{
/*				printf("P %08X 7F 128\n",count+0x38); */
				fputc(0x7F,out);
				for (bpos=0;bpos<128;bpos++)
					fputc(buff[bpos],out);
				bpos = 0;
			}
			buff[bpos] = fgetc(in);
			bpos++;
			if (bpos >= 2) /* Need at least 2 to check for a compressed segment */
			{
				if (buff[bpos-1] == buff[bpos-2])
				{
					/* Compressed segment */
					if (bpos-3 >= 0)
					{
/*						printf("P %08X %02X %d\n",count+0x38,bpos-3,bpos-3); */
						fputc(bpos-3,out);
						for (tmp=0;tmp<bpos-2;tmp++)
							fputc(buff[tmp],out);
					}
					buff[0] = buff[bpos-1];
					tmp = 1; /* compressed segment */
					bpos = 2; /* 2 pixels so far */
				}
			}
		}
		else
		{
			buff[1] = fgetc(in);
			bpos++;
			if (buff[0] != buff[1])
			{
				/* Abort compressed segment */
				if (bpos > 2)
				{
/*					printf("C %08X %02X %d %02X\n",count+0x38,bpos+125,bpos-1,buff[0]); */
					fputc(bpos+125,out);
					fputc(buff[0],out);
					buff[0] = buff[1];
					bpos = 1;
				}
				else
					bpos = 2; /* Not enough data to compress */
				tmp = 0;
			}
			else if (bpos == 129) /* Full? */
			{
/*				printf("C %08X %02X %d %02X\n",count+0x38,0xFF,129,buff[0]); */
				fputc(0xFF,out);
				fputc(buff[0],out);
				bpos = 0;
				tmp = 0; /* Start again */
			}
		}
	}
	if (bpos > 0)
	{
		if (tmp == 0)
		{
			fputc(bpos-1,out);
			for (tmp=0;tmp<bpos;tmp++)
				fputc(buff[tmp],out);
		}
		else
		{
			if (bpos == 1)
			{
				fputc(0,out);
				fputc(buff[0],out);
			}
			else
			{
				fputc(bpos+126,out);
				fputc(buff[0],out);
			}
		}
	}
	return 0xffd;
}

/*

				HAMMER CONVERTERS

*/

int32_t fromhammer()
{
	int32_t tmp;
	tmp = ihmr_getc();
	while (tmp != EOF)
	{
		fputc(tmp,out);
		tmp = ihmr_getc();
	}
	return 0xffd; /* Just do data format */
}


/*

				MAIN FUNCTION

*/

#define REDIR ((char *) 1)
#define NOTSPEC ((char *) 0)

int main(int argc,char **argv)
{
	int32_t prop = 0; /* Process op - 0=level, 1=backfile, 2=fastspr, 3=compressed background 'edge' file */
			/* 4=hammer decompress */
	int32_t dir = 0; /* 0=auto, 1=from asylum, 2=to asylum */
	char *in_name = NOTSPEC,*out_name = NOTSPEC; /* 1 to use stdin/out */
	/* Process arguments... */
	int32_t a,b;
	int32_t type;
#ifdef __riscos__
	_kernel_swi_regs regs;
#endif
	for (a=1;a<argc;a++)
	{
		if (argv[a][0] == '-') /* Switch */
		{
			if (strlen(argv[a]) > 1)
			{
				if (argv[a][1] == '-') /* int32_t switch */
				{
					if (strcmp(argv[a],"--help") == 0)
					{
						dohelp();
						exit(EXIT_SUCCESS);
					}
					else
					{
						printf("nobrainer: Unrecognised switch '%s'\n",argv[a]);
						exit(EXIT_FAILURE);
					}
				}
				else
					for (b=1;b<(int32_t) strlen(argv[a]);b++)
					{
						switch (argv[a][b])
						{
						case 'l':
							prop = 0;
							break;
						case 'b':
							prop = 1;
							break;
						case 's':
							prop = 2;
							break;
						case 'e':
							prop = 3;
							break;
						case 'm':
							prop = 4;
							break;
						case 'f':
							dir = 1;
							break;
						case 't':
							dir = 2;
							break;
						case 'h':
							dohelp();
							exit(EXIT_SUCCESS);
						default:
							fprintf(stderr,"nobrainer: Unrecognised switch '%c'\n",argv[a][b]);
							exit(EXIT_FAILURE);
						}
					}
			}
			else if (in_name == NOTSPEC)
				in_name = REDIR;
			else
				out_name = REDIR; /* Just replace it */
		}
		else if (in_name == NOTSPEC)
			in_name = argv[a];
		else
			out_name = argv[a];
	}
	if (in_name == NOTSPEC)
		in_name = REDIR;
	if (out_name == NOTSPEC)
		out_name = REDIR;
	/* Now work out whether the arguments made sense */
#ifndef __riscos__
	if (dir == 0)
	{
		fprintf(stderr,"nobrainer: Unable to guess conversion direction on non-RISC OS machines\n");
		exit(EXIT_FAILURE);
	}
#endif
	if ((in_name == REDIR) && (dir == 0))
	{
		fprintf(stderr,"nobrainer: Unable to guess conversion direction if input is not a file\n");
		exit(EXIT_FAILURE);
	}
	/* Arguments made sense */
	/* Need to identify the direction */
#ifdef __riscos__
	if (dir == 0)
	{
		regs.r[0] = 17;
		regs.r[1] = (int32_t) in_name;
		_kernel_swi(OS_File,&regs,&regs);
		type = (((uint32_t) regs.r[2]) >> 8);
		if ((type == 0xFFFFFD) || (type == 0xFFF113)) /* Data or fastspr */
			dir = 1; /* From asylum */
		else
			dir = 2; /* To asylum */
	}
#endif
	/* Now open files */
	if (in_name == REDIR)
		in = stdin;
	else
		in = fopen(in_name,"rb");
	if (in == 0)
	{
		fprintf(stderr,"nobrainer: Unable to open input file '%s'\n",in_name);
		exit(EXIT_FAILURE);
	}
	if (out_name == REDIR)
		out = stdout;
	else
		out = fopen(out_name,"wb");
	if (out == 0)
	{
		fclose(in);
		fprintf(stderr,"nobrainer: Unable to open output file '%s'\n",out_name);
		exit(EXIT_FAILURE);
	}
	/* Now process */
	type = 0;
	switch (prop+(dir*256))
	{
	case 0+256:
		type = fromlevel();
		break;
	case 0+512:
		type = tolevel();
		break;
	case 1+256:
		type = frombackfile();
		break;
	case 1+512:
		type = tobackfile();
		break;
	case 2+256:
		type = fromfastspr();
		break;
	case 2+512:
		type = tofastspr();
		break;
	case 3+256:
		type = fromedge();
		break;
	case 3+512:
		type = toedge();
		break;
	case 4+256:
		type = fromhammer();
		break;
	case 4+512:
		fprintf(stderr,"nobrainer: Unable to generate Hammer files in this version\n");
		break;
	default:
		fprintf(stderr,"nobrainer: Doh! Produced an unknown conversion code %" PRId32 " :(\n",prop+(dir*256));
	}
	fclose(in);
	fclose(out);
#ifdef __riscos__
	if ((type != 0) && (out_name != REDIR))
	{
		regs.r[0] = 18;
		regs.r[1] = (int32_t) out_name;
		regs.r[2] = type;
		_kernel_swi(OS_File,&regs,&regs);
	}
#endif
	return 0;
}



